# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
=========================
Test Framework Exceptions
=========================

This module contains the DTestException class, which is an extension
raised by the framework when an error is encountered while executing
functions or methods of the framework itself.
"""


class DTestException(Exception):
    """
    DTestException
    ==============

    The DTestException is an exception class used for all exceptions
    generated by the test framework.
    """

    pass
